//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
// Class          : Sort
//
// Author         : Richard E. Pattis
//                  Computer Science Department
//                  Carnegie Mellon University
//                  5000 Forbes Avenue
//                  Pittsburgh, PA 15213-3891
//                  e-mail: pattis@cs.cmu.edu
//
// Maintainer     : Author
//
//
// Description:
//
//   The Sort class contains a variety of sorting methods. Each has the
// same prototype, consisting of an array, its size, and an Orderable
// object. The result is a permutation of the array (same values) but
// ordered according to the Orderable object. 
//
// Future Plans   : Add more Sorts
//                  Rewrite Sorts to be as fast as possible
//
// Program History:
//   10/25/01: R. Pattis - Operational for 15-111
//    9/16/04: R. Pattis - Changed Orderable to Comparator for 15-200
//
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

import java.util.Comparator;



/** 
 * The <code>Sort</code> class is a (static) method library supplying
 *   a variety of sorting methods. Some are slow ( O(N<sup>2</sup>) ) and some are
 *   fast ( O(N Log<sub>2</sub> N) ); some are stable (equal keys retain their
 *   relative order) and some are not stable.
 * 
 * @author Richard E. Pattis (Computer Science Department, Carnegie Mellon)
*/
public class Sort {


	/** 
	 * Sorts <code>a[0:size-1]</code> according to the
	 *   <code>beongsBefore</code> method in <code>o</code>.
	 * 
	 * @param  a    specifies the array to sort
	 * @param  size specifies the number of values in the array
	 * @param  c    specifies an object whose <code>compare</code>
	 *              method determines the final order of the array.
	 * <p>
	 * Bubble sort is an unstable O(N<sup>2</sup>) method (comparisons and swaps).
	*/
	public static void bubble (Object[] a, int size, Comparator c)
	{
	  for (int base=0; base<size; base++)
	    for (int check=base+1; check<size; check++)
	      if (c.compare(a[base],a[check]) > 0)
	        swap(a,base,check);
  }
  

	
	/** 
	 * Sorts <code>a[0:size-1]</code> according to the
	 *   <code>beongsBefore</code> method in <code>o</code>.
	 * 
	 * @param  a    specifies the array to sort
	 * @param  size specifies the number of values in the array
	 * @param  c    specifies an object whose <code>compare</code>
	 *              method determines the final order of the array.
	 * <p>
	 * Selection sort is an unstable O(N<sup>2</sup>) method
	     (O(N<sup>2</sup>) comparisons and O(N) swaps).
	*/
	public static void selection (Object[] a, int size, Comparator c)
	{
    for (int passes=0; passes<size; passes++) {
      int indexOfMin = passes;
      for (int i=passes+1; i<size; i++)
        if ( c.compare(a[indexOfMin],a[i]) > 0 )
          indexOfMin = i;
      swap(a,passes,indexOfMin);
    }
  }
  


	/** 
	 * Sorts <code>a[0:size-1]</code> according to the
	 *   <code>beongsBefore</code> method in <code>o</code>.
	 * 
	 * @param  a    specifies the array to sort
	 * @param  size specifies the number of values in the array
	 * @param  c    specifies an object whose <code>compare</code>
	 *              method determines the final order of the array.
	 * <p>
	 * Insertion sort is a stable O(N<sup>2</sup>) method
	     (O(N<sup>2</sup>) comparisons swaps).
	*/
	public static void insertion (Object[] a, int size, Comparator c)
	{
    for (int passes=0; passes<size; passes++) {
      Object value = a[passes];
      for (int i=passes-1; i>=0; i--)
        if ( c.compare(value,a[i]) < 0 ) {
          a[i+1] = a[i];
          if (i == 0) {
            a[0] = value;
            break;
          }
        }else{
          a[i+1] = value;
          break;
        }
    }
  }
  

	
	/** 
	 * Sorts <code>a[0:size-1]</code> according to the
	 *   <code>beongsBefore</code> method in <code>o</code>.
	 * 
	 * @param  a    specifies the array to sort
	 * @param  size specifies the number of values in the array
	 * @param  c    specifies an object whose <code>compare</code>
	 *              method determines the final order of the array.
	 * <p>
	 * Merge sort is a stable O(N Log<sub>2</sub> N) method
	     (O(N Log<sub>2</sub> N) comparisons swaps).
	 * It allocates another array of length <code>size</code>
	*/
	public static void merge (Object[] a, int size, Comparator c)
	{mergeSort(a,0,size-1,new Object[size],c);}
  

	/** 
	 * Sorts <code>a[0:size-1]</code> according to the
	 *   <code>beongsBefore</code> method in <code>o</code>.
	 * 
	 * @param  a    specifies the array to sort
	 * @param  size specifies the number of values in the array
	 * @param  c    specifies an object whose <code>compare</code>
	 *              method determines the final order of the array.
	 * <p>
	 * Quick sort is an unstable O(N Log<sub>2</sub> N) method
	     (O(N Log<sub>2</sub> N) comparisons swaps).
	*/
	public static void quick (Object[] a, int size, Comparator c)
	{quickSort(a,0,size-1,c);}
  

	/** 
	 * Sorts <code>a[0:size-1]</code> according to the
	 *   <code>beongsBefore</code> method in <code>o</code>.
	 * 
	 * @param  a    specifies the array to sort
	 * @param  size specifies the number of values in the array
	 * @param  o    specifies an object whose <code>belongsBefore</code>
	 *              method determines the final order of the array.
	 * <p>
	 * Heap sort is an unstable O(N Log<sub>2</sub> N) method
	     (O(N Log<sub>2</sub> N) comparisons swaps).
	*/
	public static void heap (Object[] a, int size, Comparator c)
	{
	  for (int i=1; i<size; i++)
	    addToHeap(a,i,c);
	    
	  for (int i=size-1; i>=0; i--)
	    removeFromHeap(a,i,c);
	    
	  for (int i=0; i<size-1; i++)
	    if (c.compare(a[i],a[i+1]) > 0)
	      swap(a,i,i+1);
	}
  




	/** 
	 * Swaps <code>a[i]</code> with <code>a[j]</code>.
	 * 
	 * @param  a specifies the array in which to swap values
	 * @param  i specifies the index of one value to swap
	 * @param  j specifies the index of the other value to swap
	*/
  public static void swap (Object[] a, int i, int j)
  {
    Object temp = a[i];
    a[i] = a[j];
    a[j] = temp;
  }
   
   
  
  //Public helper methods

  private static void sort2 (Object[] a, int i, int j, Comparator c)
  {
    if (c.compare(a[i],a[j]) > 0)
      swap(a,i,j);
  }
   
  
  private static void sort3 (Object[] a, int i, int j, int k, Comparator c)
  {
    sort2(a,i,j,c);
    sort2(a,i,k,c);
    sort2(a,j,k,c);
  }
   
  
  private static void mergeSort(Object[]a,
                                int low,
                                int high,
                                Object[] temp,
                                Comparator c)
  {
    int size = high-low+1;
    if (size <= 1)
      return;
      
    int mid = low+size/2;
    
    //Sort parts
    mergeSort(a, low, mid-1, temp, c);
    mergeSort(a, mid, high,  temp, c);
    
    //Merge sorted parts
    for (int i=low,j=mid,t=low; t!=high+1; t++) {
      if (i == mid)
        temp[t] = a[j++];
      else if (j == high+1)
        temp[t] = a[i++];
      else if (c.compare(a[i],a[j]) < 0)
        temp[t] = a[i++];
      else
        temp[t] = a[j++];
    }
    
    //Put back into orignal array a
    for (int i=low; i<=high; i++)
      a[i] = temp[i];
  }
  
  
   
  private static void quickSort(Object[]a,
                                int left,
                                int right,
                                Comparator c)
  {
    int size = right-left+1;
    if (size <= 1)
      return;
    else if (size == 2) {
      sort2(a,left,right,c);
      return;
    }
      
    int mid = (left+right)/2;
    
    sort3(a, left, mid, right, c);
    swap(a,mid,right);
        
    //Partition it
    Object pivot = a[right];
    int    l     = left-1;
    int    r     = right;
    for (;;) {
      for(;;)
        if (c.compare(pivot,a[++l]) < 0)
          break;
      for(;;)
        if (c.compare(a[--r],pivot) < 0)
          break;
      swap(a,l,r);
      if (r <= l)
        break;
    }
    swap(a,l,r);
    swap(a,l,right);
    
    //Sort each partition
    quickSort(a, left, l-1,  c);
    quickSort(a, l+1, right, c);
  } 


	private static void addToHeap (Object[] a, int pos, Comparator c)
	{
	  for (int i=pos; i/2!=0; i=i/2)
	    if (c.compare(a[i/2],a[i]) > 0)
	      break;
	    else
	      swap(a,i,i/2);
	}


	private static void removeFromHeap (Object[] a, int pos, Comparator c)
	{
	  swap(a,1,pos);
	  pos--;
	  
		for (int i=1; 2*i<=pos;) {
			int nextI;

			//Which child to check against
			if (pos < 2*i+1 || c.compare(a[2*i],a[2*i+1]) > 0 )
			  nextI = 2*i;
			else
			  nextI = 2*i+1;

			//Check for termination or swap and proceed to next
			if ( c.compare(a[i],a[nextI]) > 0 )
			  break;
			else {
			  swap(a,i,nextI);
			  i = nextI;
			}
		}
	}

}